"""
SPOS - Small Payload Object Serializer
MIT License

Copyright (c) 2020 [Luiz Eduardo Amaral](luizamaral306@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from . import TestCase
import spos
from spos import random as srandom


class TestRandomMessageAndValue(TestCase):
    n_tests = 1000

    def evaluate_message(self, block_spec):
        block = spos.Block(block_spec)
        messages = set()
        values = list()
        for i in range(self.n_tests):
            message = srandom.block_random_message(block_spec)
            messages.add(message)
            values.append(block.bin_decode(message))

        if block.type == "pad":
            self.assertEqual(len(messages), 1)
            self.assertEqual(len(set(values)), 1)
            return

        if block.type not in ["array", "object"]:
            self.assertEqual(len(message) - 2, block.bits)

        self.assertGreaterEqual(len(messages), 2)
        self.assertGreaterEqual(len(values), 2)

    def evaluate_value(self, block_spec):
        block = spos.Block(block_spec)
        messages = set()
        values = list()
        for i in range(self.n_tests):
            value = srandom.block_random_value(block_spec)
            values.append(value)
            message = block.bin_encode(value)
            messages.add(message)

        if block.type not in ["array", "object"]:
            self.assertEqual(len(message) - 2, block.bits)

        if block.type == "pad":
            return
        self.assertGreaterEqual(len(messages), 2)
        self.assertGreaterEqual(len(values), 2)

    def test_seed(self):
        block_spec = {
            "key": "integer",
            "type": "integer",
            "bits": 16,
        }
        messages = set()
        values = list()
        block = spos.Block(block_spec)
        for i in range(self.n_tests):
            srandom.seed(1)
            message = srandom.block_random_message(block_spec)
            messages.add(message)
            values.append(block.bin_decode(message))

        if block.type not in ["array", "object"]:
            self.assertEqual(len(message) - 2, block.bits)

        self.assertEqual(len(messages), 1)
        self.assertEqual(len(set(values)), 1)

    def test_message_boolean(self):
        block_spec = {
            "key": "bool",
            "type": "boolean",
        }
        self.evaluate_message(block_spec)

    def test_message_binary(self):
        block_spec = {"key": "binary", "type": "binary", "bits": 16}
        self.evaluate_message(block_spec)

    def test_message_integer(self):
        block_spec = {"key": "integer", "type": "integer", "bits": 16}
        self.evaluate_message(block_spec)

    def test_message_float(self):
        block_spec = {"key": "float", "type": "float", "bits": 16}
        self.evaluate_message(block_spec)

    def test_message_pad(self):
        block_spec = {"key": "pad", "type": "pad", "bits": 16}
        self.evaluate_message(block_spec)

    def test_message_array(self):
        block_spec = {
            "key": "array",
            "type": "array",
            "bits": 6,
            "blocks": {"key": "float", "type": "float", "bits": 4},
        }
        self.evaluate_message(block_spec)

    def test_message_object(self):
        block_spec = {
            "key": "object",
            "type": "object",
            "blocklist": [
                {"key": "float", "type": "float", "bits": 4},
                {"key": "boolean", "type": "boolean"},
                {
                    "key": "array",
                    "type": "array",
                    "bits": 4,
                    "blocks": {"key": "integer", "type": "integer", "bits": 6},
                },
            ],
        }
        self.evaluate_message(block_spec)

    def test_message_string(self):
        block_spec = {"key": "string", "type": "string", "length": 16}
        self.evaluate_message(block_spec)

    def test_message_steps(self):
        block_spec = {
            "key": "steps",
            "type": "steps",
            "steps": [0, 1, 2, 3, 4, 5],
        }
        self.evaluate_message(block_spec)

    def test_message_categories(self):
        block_spec = {
            "key": "categories",
            "type": "categories",
            "categories": ["one", "two", "three", "four"],
        }
        self.evaluate_message(block_spec)

    def test_value_boolean(self):
        block_spec = {
            "key": "bool",
            "type": "boolean",
        }
        self.evaluate_value(block_spec)

    def test_value_binary(self):
        block_spec = {"key": "binary", "type": "binary", "bits": 16}
        self.evaluate_value(block_spec)

    def test_value_integer(self):
        block_spec = {"key": "integer", "type": "integer", "bits": 16}
        self.evaluate_value(block_spec)

    def test_value_float(self):
        block_spec = {"key": "float", "type": "float", "bits": 16}
        self.evaluate_value(block_spec)

    def test_value_pad(self):
        block_spec = {"key": "pad", "type": "pad", "bits": 16}
        self.evaluate_value(block_spec)

    def test_value_array(self):
        block_spec = {
            "key": "array",
            "type": "array",
            "bits": 6,
            "blocks": {"key": "float", "type": "float", "bits": 4},
        }
        self.evaluate_value(block_spec)

    def test_value_object(self):
        block_spec = {
            "key": "object",
            "type": "object",
            "blocklist": [
                {"key": "float", "type": "float", "bits": 4},
                {"key": "boolean", "type": "boolean"},
                {
                    "key": "array",
                    "type": "array",
                    "bits": 4,
                    "blocks": {"key": "integer", "type": "integer", "bits": 6},
                },
            ],
        }
        self.evaluate_value(block_spec)

    def test_value_string(self):
        block_spec = {"key": "string", "type": "string", "length": 16}
        self.evaluate_value(block_spec)

    def test_value_steps(self):
        block_spec = {
            "key": "steps",
            "type": "steps",
            "steps": [0, 1, 2, 3, 4, 5],
        }
        self.evaluate_value(block_spec)

    def test_value_categories(self):
        block_spec = {
            "key": "categories",
            "type": "categories",
            "categories": ["one", "two", "three", "four"],
        }
        self.evaluate_value(block_spec)


class TestRandomPayload(TestCase):
    n_tests = 1000

    def evaluate(self, payload_spec):
        messages = set()
        for i in range(self.n_tests):
            message, payload_data = srandom.random_payload(payload_spec)
            messages.add(message)
            self.assertEqual(spos.encode(payload_data, payload_spec), message)
        self.assertGreaterEqual(len(messages), 2)

    def test_random_payload_0(self):
        payload_spec = {
            "name": "test random payload",
            "version": 1,
            "meta": {
                "crc8": True,
                "header": [
                    {"key": "root", "value": 123},
                    {"key": "xp", "type": "boolean"},
                ],
            },
            "body": [
                {"key": "holy", "type": "string", "length": 10},
                {"key": "version", "type": "integer", "value": 1, "bits": 6},
                {
                    "key": "buffer",
                    "type": "array",
                    "bits": 8,
                    "blocks": {"key": "buf_val", "type": "integer", "bits": 8},
                },
                {"key": "date", "type": "float", "bits": 6},
            ],
        }
        self.evaluate(payload_spec)



"""
SPOS - Small Payload Object Serializer
MIT License

Copyright (c) 2020 [Luiz Eduardo Amaral](luizamaral306@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import json
import re
import subprocess

from spos import random as srandom
import spos
from . import TestCase, test_specs


class TestSposBin(TestCase):
    n_random = 20
    test_specs = test_specs

    def stdin_encode(self, spec, payload, output):
        proc = subprocess.run(
            ["bin/spos", "-f", output, "-p", spec],
            stdout=subprocess.PIPE,
            input=json.dumps(payload).encode("utf-8"),
        )
        return proc.stdout, proc

    def stdin_decode(self, spec, message, output):
        proc = subprocess.run(
            ["bin/spos", "-d", "-m", "-f", output, "-p", spec],
            stdout=subprocess.PIPE,
            input=message,
        )
        return proc.stdout, proc

    def random_encode(self, spec, output):
        proc = subprocess.run(
            ["bin/spos", "-r", "-f", output, "-p", spec],
            stdout=subprocess.PIPE,
        )
        return proc.stdout, proc

    def random_decode(self, spec, output):
        proc = subprocess.run(
            ["bin/spos", "-r", "-d", "-f", output, "-p", spec],
            stdout=subprocess.PIPE,
        )
        return proc.stdout, proc

    def test_random_encode(self):
        for name, payload_spec_file in self.test_specs.items():
            for i in range(self.n_random):
                with self.subTest(f"{name}_random_encode{i}"):
                    message, proc = self.random_encode(
                        payload_spec_file, "bytes"
                    )
                    self.assertEqual(proc.returncode, 0)
                    message, proc = self.random_encode(
                        payload_spec_file, "hex"
                    )
                    self.assertEqual(proc.returncode, 0)
                    message, proc = self.random_encode(
                        payload_spec_file, "bin"
                    )
                    self.assertEqual(proc.returncode, 0)

    def test_random_decode(self):
        for name, payload_spec_file in self.test_specs.items():
            for i in range(self.n_random):
                with self.subTest(f"{name}_random_decode_{i}"):
                    data, proc = self.random_decode(payload_spec_file, "bytes")
                    self.assertEqual(proc.returncode, 0)
                    data, proc = self.random_decode(payload_spec_file, "hex")
                    self.assertEqual(proc.returncode, 0)
                    data, proc = self.random_decode(payload_spec_file, "bin")
                    self.assertEqual(proc.returncode, 0)

    def test_random_encode_then_decode(self):
        for name, payload_spec_file in self.test_specs.items():
            for i in range(self.n_random):
                with self.subTest(f"{name}_random_encode_then_decode{i}"):
                    with open(payload_spec_file, "r") as fp:
                        payload_spec = json.load(fp)
                    self._test_random_encode_then_decode(
                        payload_spec, payload_spec_file, "bytes"
                    )
                    self._test_random_encode_then_decode(
                        payload_spec, payload_spec_file, "bin"
                    )
                    self._test_random_encode_then_decode(
                        payload_spec, payload_spec_file, "hex"
                    )

    def _test_random_encode_then_decode(
        self, payload_spec, payload_spec_file, fmt
    ):
        message, payload_data = srandom.random_payload(payload_spec, fmt)
        message_bin, proc = self.stdin_encode(
            payload_spec_file, payload_data, fmt
        )
        message_b = (
            message.encode("utf-8") if isinstance(message, str) else message
        )
        self.assertEqual(message_b, message_bin)
        payload_bin, proc = self.stdin_decode(
            payload_spec_file, message_b, fmt
        )
        self.assertEqual(
            spos.decode(message, payload_spec), json.loads(payload_bin),
        )

    def test_version(self):
        proc = subprocess.run(["bin/spos", "-v"], stdout=subprocess.PIPE,)
        self.assertTrue(
            re.match(
                "^spos v[0-9]+[.][0-9]+[.][0-9]+(-([ab]|rc[0-9]+))?$",
                str(proc.stdout, "utf-8").strip(),
            )
            is not None
        )


# class CandyTestCase(unittest.TestCase):
#     def testCandy(self):
#         candyOutput = candy.candy()
#
#         assert candyOutput == "candy"



"""
SPOS - Small Payload Object Serializer
MIT License

Copyright (c) 2020 [Luiz Eduardo Amaral](luizamaral306@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import warnings

from spos.blocks import truncate_bits
from . import TestCase
import spos


class TestValidateBlock(TestCase):
    def test_key_not_in_block(self):
        block = {"type": "boolean"}
        with self.assertRaises(KeyError):
            spos.Block(block)

    def test_key_value_not_a_string(self):
        block = {"type": "boolean", "key": True}
        with self.assertRaises(TypeError):
            spos.Block(block)

    def test_type_not_in_block(self):
        block = {"key": "aloha"}
        with self.assertRaises(KeyError):
            spos.Block(block)

    def test_undefined_type(self):
        block = {"key": "aloha", "type": "hakuna"}
        with self.assertRaises(ValueError):
            spos.Block(block)

    def test_missing_required_key(self):
        block = {"key": "aloha", "type": "integer"}
        with self.assertRaises(KeyError):
            spos.Block(block)

    def test_wrong_type_for_required_key(self):
        block = {"key": "aloha", "type": "integer", "bits": "mattata"}
        with self.assertRaises(TypeError):
            spos.Block(block)

    def test_wrong_type_for_optional_key(self):
        block = {
            "key": "aloha",
            "type": "integer",
            "bits": 1,
            "offset": "mattata",
        }
        with self.assertRaises(TypeError):
            spos.Block(block)

    def test_invalid_key(self):
        block = {
            "key": "aloha",
            "type": "integer",
            "bits": 1,
            "mattata": "mattata",
        }
        with self.assertRaises(KeyError):
            spos.Block(block)


class TestBlock(TestCase):
    def test_boolean_true(self):
        block = {"key": "boolean encode true", "type": "boolean"}
        t = True
        a = "0b1"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t)

    def test_boolean_false(self):
        block = {"key": "boolean false", "type": "boolean"}
        t = False
        a = "0b0"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t)

    def test_binary_block(self):
        block = {"key": "binary bin", "type": "binary", "bits": 10}
        t = "0b100101"
        a = "0b0000100101"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), a)
        self.assertEqual(spos.encode_block(a, block), a)

    def test_binary_value_error(self):
        with self.assertRaises(ValueError):
            block = {
                "key": "binary bin error",
                "type": "binary",
                "bits": 10,
            }
            spos.encode_block("0xfail", block)

    def test_binary_hex(self):
        block = {
            "key": "binary large hex",
            "type": "binary",
            "bits": 10,
        }
        t = "0xdeadbeef"
        a = bin(int(t, 16))[:12]
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), a)

    def test_binary_small_hex(self):
        block = {
            "key": "binary small hex",
            "type": "binary",
            "bits": 10,
        }
        t = "0xff"
        a = truncate_bits(bin(int(t, 16))[:12], block["bits"])
        self.assertEqual(spos.encode_block(t, block), a)

    def test_binary_hex_value_error(self):
        with self.assertRaises(ValueError):
            block = {
                "key": "binary hex fail",
                "type": "binary",
                "bits": 10,
            }
            spos.encode_block("0xfail", block)

    def test_binary_type_error(self):
        with self.assertRaises(ValueError):
            block = {
                "key": "binary type fail",
                "type": "binary",
                "bits": 10,
            }
            spos.encode_block("fail", block)

    def test_integer_block(self):
        block = {
            "key": "integer test",
            "type": "integer",
            "bits": 6,
        }
        t = 1
        a = "0b000001"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t)

    def test_integer_type_error(self):
        with self.assertRaises(TypeError):
            block = {
                "key": "integer fail",
                "type": "integer",
                "bits": 6,
            }
            spos.encode_block("fail", block)

    def test_integer_offset(self):
        block = {
            "key": "integer offset",
            "type": "integer",
            "bits": 6,
            "offset": 100,
        }
        t = 120
        a = "0b010100"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t)

    def test_integer_underflow(self):
        block = {
            "key": "integer underflow",
            "type": "integer",
            "bits": 6,
        }
        t = -10
        a = "0b000000"
        self.assertEqual(spos.encode_block(t, block), a)

    def test_integer_overflow(self):
        block = {
            "key": "integer overflow",
            "type": "integer",
            "bits": 6,
        }
        t = 128
        a = "0b111111"
        self.assertEqual(spos.encode_block(t, block), a)

    def test_float_block(self):
        block = {"key": "float test", "type": "float", "bits": 2}
        t = 0.66
        a = "0b10"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertClose(spos.decode_block(a, block), t)

    def test_float_type_error(self):
        with self.assertRaises(TypeError):
            block = {
                "key": "float fail",
                "type": "float",
                "bits": 4,
            }
            spos.encode_block("fail", block)

    def test_float_approximation_floor(self):
        block = {
            "key": "float approximation floor",
            "type": "float",
            "bits": 2,
            "approximation": "floor",
        }
        t = 0.66
        a = "0b01"
        t_dec = 0.33
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertClose(spos.decode_block(a, block), t_dec)

    def test_float_approximation_ceil(self):
        block = {
            "key": "float approximation ceil",
            "type": "float",
            "bits": 2,
            "approximation": "ceil",
        }
        t = 0.34
        a = "0b10"
        t_dec = 0.66
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertClose(spos.decode_block(a, block), t_dec)

    def test_float_upper(self):
        block = {
            "key": "float upper",
            "type": "float",
            "bits": 5,
            "upper": 31,
        }
        t = 13
        a = "0b01101"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertClose(spos.decode_block(a, block), t)

    def test_float_lower(self):
        block = {
            "key": "float lower",
            "type": "float",
            "bits": 3,
            "lower": -6,
        }
        t = -1
        a = "0b101"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertClose(spos.decode_block(a, block), t)

    def test_float_underflow(self):
        block = {
            "key": "float underflow",
            "type": "float",
            "bits": 6,
        }
        t = -10
        a = "0b000000"
        self.assertEqual(spos.encode_block(t, block), a)

    def test_float_overflow(self):
        block = {
            "key": "float overflow",
            "type": "float",
            "bits": 6,
        }
        t = 10
        a = "0b111111"
        self.assertEqual(spos.encode_block(t, block), a)

    def test_pad_block_2(self):
        block = {"key": "pad test", "type": "pad", "bits": 2}
        t = None
        a = "0b11"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t)

    def test_pad_block_6(self):
        block = {"key": "pad test", "type": "pad", "bits": 6}
        t = None
        a = "0b111111"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t)

    def test_array_block(self):
        block = {
            "key": "integer array",
            "type": "array",
            "bits": 3,
            "blocks": {"key": "array val", "type": "integer", "bits": 3},
        }
        t = [1, 2, 3, 4]
        a = "0b100001010011100"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t)

    def test_array_truncate(self):
        block = {
            "key": "truncate array",
            "type": "array",
            "bits": 2,
            "blocks": {"key": "array val", "type": "integer", "bits": 3},
        }
        t = [1, 2, 3, 4, 5]
        a = "0b11001010011"
        t_dec = [1, 2, 3]
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t_dec)

    def test_array_empty(self):
        block = {
            "key": "empty array",
            "type": "array",
            "bits": 7,
            "blocks": {"key": "array val", "type": "integer", "bits": 3},
        }
        t = []
        a = "0b0000000"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), [])

    def test_array_nested(self):
        block = {
            "key": "nested array",
            "type": "array",
            "bits": 2,
            "blocks": {
                "key": "array val 1",
                "type": "array",
                "bits": 3,
                "blocks": {"key": "array val 2", "type": "integer", "bits": 3},
            },
        }
        t = [[1, 2], [3, 4, 5]]
        a = "0b10010001010011011100101"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t)

    def test_object_block(self):
        block = {
            "key": "object",
            "type": "object",
            "blocklist": [
                {"key": "key1", "type": "integer", "bits": 3},
                {"key": "key2", "type": "float", "bits": 5},
            ],
        }
        t = {"key1": 6, "key2": 0.9}
        a = "0b11011100"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertDict(spos.decode_block(a, block), t)

    def test_object_nested(self):
        block = {
            "key": "object",
            "type": "object",
            "blocklist": [
                {"key": "key1", "type": "integer", "bits": 3},
                {
                    "key": "key2",
                    "type": "object",
                    "blocklist": [
                        {"key": "nKey", "type": "boolean"},
                        {
                            "key": "nKey2",
                            "type": "object",
                            "blocklist": [
                                {"key": "nKey3", "type": "float", "bits": 8}
                            ],
                        },
                    ],
                },
            ],
        }
        t = {"key1": 6, "key2": {"nKey": False, "nKey2": {"nKey3": 0.8}}}
        a = "0b110011001100"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertDict(spos.decode_block(a, block), t)

    def test_object_dot_notation_0(self):
        block = {
            "key": "object",
            "type": "object",
            "blocklist": [
                {"key": "key1", "type": "integer", "bits": 3},
                {"key": "key2.surprise", "type": "string", "length": 5},
                {
                    "key": "key2",
                    "type": "object",
                    "blocklist": [
                        {"key": "nKey", "type": "boolean"},
                        {
                            "key": "nKey2",
                            "type": "object",
                            "blocklist": [
                                {"key": "nKey3", "type": "float", "bits": 8}
                            ],
                        },
                    ],
                },
            ],
        }
        t = {
            "key1": 6,
            "key2": {
                "surprise": "hello",
                "nKey": False,
                "nKey2": {"nKey3": 0.8},
            },
        }
        a = "0b110100001011110100101100101101000011001100"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertDict(spos.decode_block(a, block), t)

    def test_object_dot_notation_1(self):
        block = {
            "key": "object",
            "type": "object",
            "blocklist": [
                {"key": "key1.nest1.a", "type": "integer", "bits": 3},
                {"key": "key1.nest1.b", "type": "string", "length": 5},
                {"key": "key1.nest1.c", "type": "integer", "bits": 3},
                {
                    "key": "key1",
                    "type": "object",
                    "blocklist": [
                        {"key": "nest1.d", "type": "integer", "bits": 3},
                        {
                            "key": "nest1",
                            "type": "object",
                            "blocklist": [
                                {"key": "e", "type": "integer", "bits": 3},
                                {"key": "f", "type": "integer", "bits": 3},
                            ],
                        },
                    ],
                },
            ],
        }
        t = {
            "key1": {
                "nest1": {"a": 1, "b": "hello", "c": 2, "d": 3, "e": 4, "f": 5}
            },
        }
        a = "0b001100001011110100101100101101000010011100101"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertDict(spos.decode_block(a, block), t)

    def test_object_key_error(self):
        with self.assertRaises(KeyError):
            block = {
                "key": "object missing key",
                "type": "object",
                "blocklist": [
                    {"key": "key1", "type": "integer", "bits": 3},
                    {"key": "key2", "type": "float", "bits": 5},
                    {"key": "key3", "type": "boolean"},
                ],
            }
            t = {"key1": 6, "key2": 0.9}
            spos.encode_block(t, block)

    def test_string_block(self):
        block = {
            "key": "string",
            "type": "string",
            "length": 6,
        }
        t = "test"
        a = "0b111110111110101101011110101100101101"
        t_dec = "++test"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t_dec)

    def test_string_block_unknown_character(self):
        block = {
            "key": "string",
            "type": "string",
            "length": 6,
        }
        t = "test%"
        a = "0b111110101101011110101100101101111111"
        t_dec = "+test/"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t_dec)

    def test_string_custom_alphabeth(self):
        block = {
            "key": "string",
            "type": "string",
            "length": 6,
            "custom_alphabeth": {62: " "},
        }
        t = "test"
        a = "0b111110111110101101011110101100101101"
        t_dec = "  test"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t_dec)

    def test_steps_block(self):
        block = {
            "key": "steps",
            "type": "steps",
            "steps": [0, 5, 10],
            "steps_names": ["critical", "low", "charged", "full"],
        }
        t = 2
        a = "0b01"
        t_dec = "low"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t_dec)

    def test_steps_index_0(self):
        block = {
            "key": "steps",
            "type": "steps",
            "steps": [0, 5, 10],
            "steps_names": ["critical", "low", "charged", "full"],
        }
        t = -1
        a = "0b00"
        t_dec = "critical"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t_dec)

    def test_steps_lower_boundary(self):
        block = {
            "key": "steps",
            "type": "steps",
            "steps": [0, 5, 10],
            "steps_names": ["critical", "low", "charged", "full"],
        }
        t = 5
        a = "0b10"
        t_dec = "charged"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t_dec)

    def test_steps_last_index(self):
        block = {
            "key": "steps",
            "type": "steps",
            "steps": [0, 5, 10],
            "steps_names": ["critical", "low", "charged", "full"],
        }
        t = 11
        a = "0b11"
        t_dec = "full"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t_dec)

    def test_steps_auto_steps_names(self):
        block = {
            "key": "steps",
            "type": "steps",
            "steps": [0, 5, 10],
        }
        t = 1
        a = "0b01"
        t_dec = "0<=x<5"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t_dec)

    def test_steps_error_steps_names(self):
        block = {
            "key": "steps",
            "type": "steps",
            "steps": [0, 5, 10],
            "steps_names": ["one", "two"],
        }
        t = 1
        a = "0b10"
        with self.assertRaises(ValueError):
            spos.encode_block(t, block)
        with self.assertRaises(ValueError):
            spos.decode_block(a, block)

    def test_steps_error_unordered_steps(self):
        block = {
            "key": "steps",
            "type": "steps",
            "steps": [5, 0, 10],
            "steps_names": ["one", "two"],
        }
        t = 1
        a = "0b10"
        with self.assertRaises(ValueError):
            spos.encode_block(t, block)
        with self.assertRaises(ValueError):
            spos.decode_block(a, block)

    def test_categories_block_0(self):
        block = {
            "key": "categories",
            "type": "categories",
            "categories": ["critical", "low", "charged", "full"],
        }
        t = "critical"
        a = "0b000"
        t_dec = "critical"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t_dec)

    def test_categories_block_1(self):
        block = {
            "key": "categories",
            "type": "categories",
            "categories": ["critical", "low", "charged", "full"],
        }
        t = "low"
        a = "0b001"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t)

    def test_categories_block_2(self):
        block = {
            "key": "categories",
            "type": "categories",
            "categories": ["critical", "low", "charged", "full"],
        }
        t = "charged"
        a = "0b010"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t)

    def test_categories_block_3(self):
        block = {
            "key": "categories",
            "type": "categories",
            "categories": ["critical", "low", "charged", "full"],
        }
        t = "full"
        a = "0b011"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t)

    def test_categories_block_4(self):
        block = {
            "key": "categories",
            "type": "categories",
            "categories": ["critical", "low", "charged", "full"],
        }
        t = "unknown"
        a = "0b100"
        t_dec = "unknown"
        self.assertEqual(spos.encode_block(t, block), a)
        self.assertEqual(spos.decode_block(a, block), t_dec)

    def test_categories_block_overflow(self):
        block = {
            "key": "categories",
            "type": "categories",
            "categories": ["critical", "low", "charged", "full"],
        }
        a = "0b111"
        t_dec = "error"
        self.assertEqual(spos.decode_block(a, block), t_dec)

    def test_crc_bin(self):
        t = "0b1011110010110010"
        a = "0b10100100"
        b = "0b101111001011001010100100"
        self.assertEqual(spos.create_crc8(t), a)
        self.assertEqual(spos.check_crc8(b), True)

    def test_crc_hex(self):
        t = "0xABCD35"
        a = "0b00101011"
        b = "0xABCD352B"
        self.assertEqual(spos.create_crc8(t), a)
        self.assertEqual(spos.check_crc8(b), True)

    def test_validate_encode_input_types_decorator_keyword_argument(self):
        block = {"key": "boolean encode true", "type": "boolean"}
        block = spos.Block(block)
        block.bin_encode(value=True)

    def test_validate_encode_input_types_decorator_type_error(self):
        block = {"key": "boolean encode true", "type": "boolean"}
        block = spos.Block(block)
        with self.assertRaises(TypeError):
            block.bin_encode(True, "err")
        with self.assertRaises(TypeError):
            block.bin_encode()

    def test_static_value(self):
        static_value = "0b10101"
        block = {
            "key": "binary bin",
            "type": "binary",
            "value": static_value,
            "bits": 5,
        }
        self.assertEqual(spos.encode_block(None, block), static_value)
        self.assertEqual(spos.encode_block("0b11", block), static_value)
        self.assertEqual(spos.encode_block(static_value, block), static_value)

    def test_static_value_mismatch_warning(self):
        static_value = "0b10101"
        block = {
            "key": "binary bin",
            "type": "binary",
            "value": static_value,
            "bits": 5,
        }
        with warnings.catch_warnings(record=True) as w:
            self.assertEqual(spos.decode_block("0b11111", block), static_value)
            self.assertEqual(len(w), 1)
            self.assertTrue(
                issubclass(w[-1].category, spos.StaticValueMismatchWarning)
            )



"""
SPOS - Small Payload Object Serializer
MIT License

Copyright (c) 2020 [Luiz Eduardo Amaral](luizamaral306@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import random

from . import utils, encode
from .blocks import Block


def seed(a=None, version=2):
    """
    Alias for random.seed
    """
    random.seed(a, version)


class RandomBlock(Block):
    def __new__(cls, *args, **kwargs):
        random_encoders = {
            "array": cls.array_random_message,
            "object": cls.object_random_message,
            "steps": cls.steps_random_message,
            "categories": cls.categories_random_message,
        }
        random_decoders = {
            "pad": cls.pad_random_value,
            "object": cls.object_random_value,
            "steps": cls.steps_random_value,
            "categories": cls.categories_random_value,
        }
        instance = super().__new__(cls, *args, **kwargs)
        setattr(
            instance,
            "random_message",
            cls.random_message.__get__(instance, instance.__class__),
        )
        setattr(
            instance,
            "_random_message",
            cls._random_message.__get__(instance, instance.__class__),
        )
        setattr(
            instance,
            "random_value",
            cls.random_value.__get__(instance, instance.__class__),
        )
        setattr(
            instance,
            "_random_value",
            cls._random_value.__get__(instance, instance.__class__),
        )
        for name, fn in random_encoders.items():
            if instance.type == name:
                setattr(
                    instance,
                    "_random_message",
                    fn.__get__(instance, instance.__class__),
                )
        for name, fn in random_decoders.items():
            if instance.type == name:
                setattr(
                    instance,
                    "_random_value",
                    fn.__get__(instance, instance.__class__),
                )
        return instance

    def random_message(self):
        """
        Creates a random message
        """
        return (
            self._random_message()
            if not self.cache_message
            else self.cache_message
        )

    def _random_message(self):
        return utils.random_bits(self.bits)

    def random_value(self):
        """
        Creates a random value
        """
        return (
            self._random_value() if not self.cache_value else self.cache_value
        )

    def _random_value(self):
        return self.bin_decode(self.random_message())

    def pad_random_value(self):
        return self._bin_encode(None)

    def array_random_message(self):
        length = random.randint(0, self.max_length)
        len_message = self.length_block.bin_encode(length)
        random_blocks = RandomBlock(self.blocks.block_spec)
        return f"{len_message}{''.join([random_blocks.random_message()[2:] for _ in range(length)])}"

    def object_random_message(self):
        message = "0b"
        for block in self.blocklist:
            random_block = RandomBlock(block.block_spec)
            message += random_block.random_message()[2:]
        return message

    def object_random_value(self):
        obj = {}
        for block in self.blocklist:
            random_block = RandomBlock(block.block_spec)
            obj[block.key] = random_block.random_value()
        obj = utils.nest_keys(obj)
        return obj

    def steps_random_value(self):
        steps = self.steps + [self.steps[0] - 1]
        return random.choice(steps)

    def steps_random_message(self):
        return self.bin_encode(self.random_value())

    def categories_random_value(self):
        return random.choice(self.categories[:-1])

    def categories_random_message(self):
        return self.bin_encode(self.random_value())


def block_random_value(block_spec):
    """
    Generates a random value within block specification.
    Returns the static value if it's defined in the block.

    Args:
        block_spec (dict): Block specification.

    Returns:
        value: Random value.
    """
    return RandomBlock(block_spec).random_value()


def block_random_message(block_spec):
    """
    Generates a random message within block specification.
    Returns the static value message if it's defined in the block.

    Args:
        block_spec (dict): Block specification.
    Returns:
        message: Random message.
    """
    return RandomBlock(block_spec).random_message()


def random_payload(payload_spec, output="bin"):
    """
    Builds a random message conforming to `payload_spec`.

    Args:
        payload_spec (dict): Payload specification.
        output (str): Output format (bin, hex or bytes). default: "bin".

    Returns:
        message (bin | hex | bytes): Random message
        payload_data (object): Equivalent payload_data to generate
            message.
    """
    utils.validate_payload_spec(payload_spec)

    meta = block_random_value(
        {
            "key": "meta",
            "type": "object",
            "blocklist": [
                block_spec
                for block_spec in payload_spec.get("meta", {}).get(
                    "header", []
                )
                if "value" not in block_spec
            ],
        }
    )
    body = block_random_value(
        {
            "key": "body",
            "type": "object",
            "blocklist": payload_spec.get("body", []),
        }
    )
    payload_data = utils.merge_dicts(meta, body)
    message = encode(payload_data, payload_spec, output)
    return message, payload_data



"""
SPOS - Small Payload Object Serializer
MIT License

Copyright (c) 2020 [Luiz Eduardo Amaral](luizamaral306@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import collections
import copy
import random

from .exceptions import SpecsVersionError


def truncate_bits(bit_str, bits):
    """
    Truncates the `bit_str` to up to `bits`.
    Args:
        bit_str (str): Bit string.
        bits (int): Number of bits to truncate.

    Returns:
        trunc_bit_str (str): Truncated bit string.
    """
    return "0b" + "0" * (bits - len(bit_str) + 2) + bit_str[2 : bits + 2]


def random_bits(n):
    """
    Returns a binary string with `n` bits.

    Args:
        n (int): Number of bits.

    Returns:
        bits (str): Binary string of random bits.
    """
    return f"0b{''.join([str(random.randint(0, 1)) for _ in range(n)])}"


def validate_payload_spec(payload_spec):
    """
    Checks payload_spec for valid keys and values. Does not validates
    blocks.

    Args:
        payload_spec (dict): Payload specification.

    Raises:
        ValueError, KeyError, TypeError: When `payload_spec` does not
            follows correct specifications. See https://github.com/luxedo/SPOS
    """
    extra_keys = [
        key
        for key in payload_spec.keys()
        if key not in ["name", "version", "body", "meta"]
    ]
    if any(extra_keys):
        raise ValueError(
            f"Found unexpected keys {extra_keys} in payload_spec."
        )
    required_keys = {"name": str, "version": int, "body": list}
    for key, tp in required_keys.items():
        if key not in payload_spec:
            raise KeyError(
                f"payload_spec {payload_spec} must have key '{key}'"
            )
        if not isinstance(payload_spec[key], tp):
            raise TypeError(
                f"payload_spec {payload_spec['name']} key '{key}' must be of type '{tp}'"
            )

    dup_keys_body = duplicate_keys(payload_spec["body"])
    if any(dup_keys_body):
        raise KeyError(f"Duplicate keys found in body: {dup_keys_body}.")

    meta = payload_spec.get("meta", {})
    if not isinstance(meta, dict):
        raise ValueError(f"meta key expected to be a dict, got {type(meta)}.")
    if meta.get("encode_version"):
        if not isinstance(meta["encode_version"], bool):
            raise TypeError(
                f"meta.encode_version expected to be boolean, got {type(meta['encode_version'])}."
            )
        if not meta.get("version_bits"):
            raise KeyError("Missing key meta.version_bits.")
        if not isinstance(meta["version_bits"], int):
            raise TypeError(
                "meta.version_bits expected to be integer, got {type(meta['version_bits'])}."
            )
        if payload_spec["version"] >= 2 ** meta["version_bits"]:
            raise ValueError(
                f"Version overflow: {payload_spec['version']} >= {2**meta['version_bits']}"
            )

    header_bl = meta.get("header", [])
    validate_header(header_bl)
    dup_keys_head = duplicate_keys(header_bl)
    if any(dup_keys_head):
        raise KeyError(f"Duplicate keys found in header: {dup_keys_head}.")

    if (
        len(header_bl)
        + len(payload_spec["body"])
        + meta.get("encode_version", False)
        == 0
    ):
        raise ValueError(
            "Payload specification does not contain data to send."
        )


def duplicate_keys(blocklist):
    """
    Checks if `blocklist` has duplicate keys.

    Args:
        blocklist

    Returns:
        dup_keys (list): List of duplicate keys.
    """
    keys = flattened_keys(blocklist)
    return [
        key for key, count in collections.Counter(keys).items() if count > 1
    ]


def flattened_keys(blocklist):
    """
    Flatten 'key' values in blocklist to dot notation. Eg:
    block["key1"]["key2"] => block["key1.key2"]

    Args:
        blocklist

    Returns
        keys (list): List of keys.
    """
    keys = []
    for block_spec in blocklist:
        if block_spec.get("type") == "object":
            nested = copy.deepcopy(block_spec.get("blocklist", []))
            for nested_spec in nested:
                nested_spec[
                    "key"
                ] = f"{block_spec['key']}.{nested_spec['key']}"
            keys.extend(flattened_keys(nested))
        else:
            keys.append(block_spec["key"])
    return keys


def nest_keys(obj):
    """
    Nests keys in obj from dot notation. Eg:
    obj["key1.key2"] => obj["key1"]["key2"]

    Args:
        obj (dict)

    Returns
        obj_nested (dict)
    """
    new_obj = {}
    for key, value in obj.items():
        if isinstance(value, dict):
            value = nest_keys(value)
        if "." in key:
            keyp = key.split(".")[0]
            keyr = ".".join(key.split(".")[1:])
            value = nest_keys({keyr: value})
            if keyp in new_obj:
                new_obj[keyp] = merge_dicts(new_obj[keyp], value)
            else:
                new_obj[keyp] = value
        else:
            if key in new_obj:
                new_obj[key] = merge_dicts(new_obj[key], value)
            else:
                new_obj[key] = value
    return new_obj


def merge_dicts(dict1, dict2):
    """
    Merges dictionaries including nested values.

    Args:
        dict1
        dict2

    Returns:
        merged_dict
    """
    merged_dict = copy.deepcopy(dict1)
    for key, value in dict2.items():
        if key not in merged_dict:
            merged_dict[key] = value
        else:
            merged_dict[key] = merge_dicts(merged_dict[key], value)
    return merged_dict


def validate_header(header_bl):
    """
    Run checks for static value in the header block.

    Args:
        header_bl (list): List of blocks for the header.

    Raises:
        KeyError: If header block does not contain 'key'.
        KeyError: If static header block does not contain 'value'.
    """
    header_static = [
        block_spec for block_spec in header_bl if "value" in block_spec
    ]

    for block in header_static:
        if "key" not in block:
            raise KeyError(f"Static block {block} must have key 'key'")
        if len(block.keys()) > 2:
            raise KeyError(
                f"Static block {block['key']} must have onlye two keys: 'key' and 'value'"
            )


def validate_specs(specs, match_versions=False):
    """
    Validates an array of specs.

    Args:
        specs (list): Array of specs.
        match_versions (bool): If True raises SpecsVersionError
            if the array has conflicting versions or different
            version_bits.

    Raises:
        SpecsVersionError: When `match_versions` is set to True and
            there's duplicate versions or name mismatch.
        Other exceptions: Listed in spos.utils.validate_payload_spec.
    """
    for payload_spec in specs:
        validate_payload_spec(payload_spec)
    if match_versions:
        versions = []
        name = None
        for payload_spec in specs:
            name = payload_spec["name"] if name is None else name
            if payload_spec["version"] in versions:
                raise SpecsVersionError(
                    f"Duplicated version in 'specs': {payload_spec['version']}."
                )
            if name != payload_spec["name"]:
                raise SpecsVersionError("Name mismatch in 'specs'.")
            versions.append(payload_spec["version"])


def remove_null_values(d):
    """
    Removes None from dictionary `d` recursively.

    Args:
        d (dict)

    Returns:
        d (dict): Dictionary without None values.
    """
    bd = copy.deepcopy(d)
    del_keys = {key for key, value in bd.items() if value is None}
    for key in del_keys:
        del bd[key]
    for key in [key for key, value in bd.items() if isinstance(value, dict)]:
        bd[key] = remove_null_values(bd[key])
    return bd



"""
SPOS - Small Payload Object Serializer
MIT License

Copyright (c) 2020 [Luiz Eduardo Amaral](luizamaral306@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import argparse
import io
import json
import random
import re
import sys

from . import encode, decode, decode_from_specs, stats, __version__
from .random import random_payload


def read_and_close_json(buf):
    d = json.load(buf)
    buf.close()
    return d


def _encode(_input, output, payload_spec, fmt):
    payload_data = read_and_close_json(_input)
    message = encode(payload_data, payload_spec, fmt)
    message = message if fmt == "bytes" else bytes(message, encoding="ascii")
    if hasattr(output, "name") and output.name == "<stdout>":
        sys.stdout.buffer.write(message)
    else:
        output.write(message)
    output.close()


def _decode(_input, output, payload_specs, fmt, show_meta):
    if hasattr(_input, "name") and _input.name == "<stdin>":
        message = sys.stdin.buffer.read()
    else:
        message = _input.read()
        _input.close()

    if fmt != "bytes":
        message = message.decode("ascii")
        if fmt == "hex":
            message = (
                message
                if re.match("^0[xX][0-9a-fA-F]+$", message)
                else f"0x{message}"
            )
        elif fmt == "bin":
            message = (
                message if re.match("^0b[01]+$", message) else f"0b{message}"
            )
    if len(payload_specs) == 1:
        decoded = decode(message, payload_specs[0])
    else:
        decoded = decode_from_specs(message, payload_specs)
    if show_meta:
        output.write(
            bytes(json.dumps(decoded, indent=2) + "\n", encoding="utf-8")
        )
    else:
        output.write(
            bytes(
                json.dumps(decoded["body"], indent=2) + "\n", encoding="utf-8"
            )
        )
    output.close()


def parse(argv=None):
    parser = argparse.ArgumentParser(
        description="Spos is a tool for serializing objects."
    )

    parser.add_argument(
        "-d", "--decode", action="store_true", help="decodes a message."
    )
    parser.add_argument(
        "-p",
        "--payload-specs",
        metavar="PAYLOAD_SPEC",
        type=argparse.FileType("r"),
        nargs="+",
        required=True,
        help="json file payload specifications.",
    )
    parser.add_argument(
        "-f",
        "--format",
        choices=["bin", "hex", "bytes"],
        default="bytes",
        help="Output format",
    )
    random_group = parser.add_mutually_exclusive_group()
    random_group.add_argument(
        "-r",
        "--random",
        action="store_true",
        help="Creates a random message/decoded_message",
    )
    random_group.add_argument(
        "-I",
        "--random-input",
        action="store_true",
        help="Creates a random payload data input",
    )
    parser.add_argument(
        "-m",
        "--meta",
        action="store_true",
        help="Outputs the metadata when decoding",
    )
    parser.add_argument(
        "-s",
        "--stats",
        action="store_true",
        help="Returns payload spec statistics",
    )
    parser.add_argument(
        "-i",
        "--input",
        nargs="?",
        type=argparse.FileType("rb"),
        default=sys.stdin.buffer,
        help="Input file",
    )
    parser.add_argument(
        "-o",
        "--output",
        nargs="?",
        type=argparse.FileType("wb"),
        default=sys.stdout.buffer,
        help="Output file",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"spos v{__version__}",
        help="Software version",
    )

    return parser.parse_args(argv)


def main(args):
    payload_specs = [
        read_and_close_json(payload_spec)
        for payload_spec in args.payload_specs
    ]

    if args.stats:
        sys.stdout.write(
            json.dumps([stats(ps) for ps in payload_specs], indent=2) + "\n"
        )
        return 0

    if args.random or args.random_input:
        args.input.detach()
        payload_specs = [random.choice(payload_specs)]
        message, payload_data = random_payload(
            payload_specs[0], output=args.format
        )

        if args.random_input:
            sys.stdout.write(json.dumps(payload_data, indent=2))
            return 0

        if args.decode:
            message = (
                message
                if isinstance(message, bytes)
                else bytes(message, encoding="ascii")
            )
            args.input = io.BytesIO(message)
        else:
            args.input = io.BytesIO(
                bytes(json.dumps(payload_data), encoding="ascii")
            )

    if args.decode:
        _decode(args.input, args.output, payload_specs, args.format, args.meta)
    else:
        if len(payload_specs) > 1:
            sys.stdout.write(
                "Specify only one specification (-p, --payload-spec) for encoding"
            )
            sys.exit(2)
        _encode(args.input, args.output, payload_specs[0], args.format)
    return 0



"""
SPOS - Small Payload Object Serializer
MIT License

Copyright (c) 2020 [Luiz Eduardo Amaral](luizamaral306@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import spos
from . import TestCase


class TestValidatePayloadSpec(TestCase):
    def test_missing_version_key_error(self):
        payload_spec = {
            "name": "john",
            "body": [{"key": "jon", "type": "bool"}],
        }
        with self.assertRaises(KeyError):
            spos.bin_encode({"jon", False}, payload_spec)

    def test_missing_name_key_error(self):
        payload_spec = {
            "version": 1,
            "body": [{"key": "jon", "type": "bool"}],
        }
        with self.assertRaises(KeyError):
            spos.bin_encode({"jon", False}, payload_spec)

    def test_missing_body_key_error(self):
        payload_spec = {
            "name": "john",
            "version": 1,
        }
        with self.assertRaises(KeyError):
            spos.bin_encode({"jon", False}, payload_spec)

    def test_version_value_type_error(self):
        payload_spec = {
            "name": "john",
            "version": "wrong type",
            "body": [{"key": "jon", "type": "boolean"}],
        }
        with self.assertRaises(TypeError):
            spos.bin_encode({"jon": False}, payload_spec)

    def test_name_value_type_error(self):
        payload_spec = {
            "name": 1,
            "version": 1,
            "body": [{"key": "jon", "type": "boolean"}],
        }
        with self.assertRaises(TypeError):
            spos.bin_encode({"jon": False}, payload_spec)

    def test_body_value_type_error(self):
        payload_spec = {
            "name": "john",
            "version": 1,
            "body": {"key": "jon", "type": "boolean"},
        }
        with self.assertRaises(TypeError):
            spos.bin_encode({"jon": False}, payload_spec)

    def test_body_dupliacte_key_error(self):
        payload_spec = {
            "name": "john",
            "version": 1,
            "body": [
                {"key": "jon", "type": "boolean"},
                {"key": "jon", "type": "binary", "bits": 10},
            ],
        }
        with self.assertRaises(KeyError):
            spos.bin_encode({"jon": False}, payload_spec)

    def test_unexpected_keys(self):
        payload_spec = {
            "name": "john",
            "version": 1,
            "body": [{"key": "jon", "type": "boolean"}],
            "extra": "key",
        }
        with self.assertRaises(ValueError):
            spos.bin_encode({"jon": False}, payload_spec)

    def test_colliding_nested_keys(self):
        payload_spec = {
            "name": "john",
            "version": 1,
            "body": [
                {"key": "jon.fearless", "type": "boolean"},
                {
                    "key": "jon",
                    "type": "object",
                    "blocklist": [
                        {"key": "fearless", "type": "integer", "bits": 10}
                    ],
                },
            ],
        }
        with self.assertRaises(KeyError):
            spos.bin_encode({"jon": {"fearless": False}}, payload_spec)

    def test_empty_payload_spec_value_error(self):
        payload_spec = {
            "name": "john",
            "version": 1,
            "body": [],
        }
        with self.assertRaises(ValueError):
            spos.bin_encode({}, payload_spec)


class TestMeta(TestCase):
    def test_meta_wrong_type(self):
        payload_spec = {
            "name": "john",
            "version": 3,
            "meta": "error",
            "body": [{"key": "jon", "type": "boolean"}],
        }
        with self.assertRaises(ValueError):
            spos.bin_encode({"sensor_name": "abc", "jon": False}, payload_spec)

    def test_header(self):
        payload_spec = {
            "name": "john",
            "version": 3,
            "meta": {
                "header": [
                    {"key": "sensor_name", "type": "string", "length": 6}
                ]
            },
            "body": [{"key": "jon", "type": "boolean"}],
        }
        payload_data = {"sensor_name": "abc", "jon": False}
        enc = spos.bin_encode(payload_data, payload_spec)
        hex_enc = spos.encode(payload_data, payload_spec, "hex")

        dec = spos.bin_decode(enc, payload_spec)
        dec, dec_meta = dec["body"], dec["meta"]
        self.assertDict(dec, {"jon": False})
        self.assertDict(
            dec_meta,
            {
                "name": "john",
                "version": 3,
                "message": hex_enc,
                "header": {"sensor_name": "+++abc"},
            },
        )

    def test_header_key_collision_error(self):
        payload_spec = {
            "name": "john",
            "version": 3,
            "meta": {
                "header": [
                    {"key": "key1", "type": "string", "length": 6},
                    {"key": "key1", "type": "integer", "bits": 12},
                ]
            },
            "body": [{"key": "jon", "type": "boolean"}],
        }

        with self.assertRaises(KeyError):
            spos.bin_encode({"key1": "abc", "jon": False}, payload_spec)

    def test_encode_version(self):
        payload_spec = {
            "name": "john",
            "version": 3,
            "meta": {"encode_version": True, "version_bits": 4},
            "body": [{"key": "jon", "type": "boolean"}],
        }
        payload_data = {"sensor_name": "abc", "jon": False}
        enc = spos.bin_encode(payload_data, payload_spec)
        hex_enc = spos.encode(payload_data, payload_spec, "hex")
        dec = spos.bin_decode(enc, payload_spec)
        dec, dec_meta = dec["body"], dec["meta"]
        self.assertDict(dec, {"jon": False})
        self.assertDict(
            dec_meta, {"name": "john", "version": 3, "message": hex_enc},
        )

    def test_encode_version_type_error(self):
        payload_spec = {
            "name": "john",
            "version": 3,
            "meta": {"encode_version": "error", "version_bits": 4},
            "body": [{"key": "jon", "type": "boolean"}],
        }
        with self.assertRaises(TypeError):
            spos.bin_encode({"sensor_name": "abc", "jon": False}, payload_spec)

    def test_version_bits_type_error(self):
        payload_spec = {
            "name": "john",
            "version": 3,
            "meta": {"encode_version": True, "version_bits": "error"},
            "body": [{"key": "jon", "type": "boolean"}],
        }
        with self.assertRaises(TypeError):
            spos.bin_encode({"sensor_name": "abc", "jon": False}, payload_spec)

    def test_encode_version_mismatch_error(self):
        payload_spec = {
            "name": "john",
            "version": 3,
            "meta": {"encode_version": True, "version_bits": 4},
            "body": [{"key": "jon", "type": "boolean"}],
        }
        encoded = "0b00100000"
        with self.assertRaises(spos.VersionError):
            spos.bin_decode(encoded, payload_spec)

    def test_encode_version_missing_version_bits(self):
        payload_spec = {
            "name": "john",
            "version": 3,
            "meta": {"encode_version": True},
            "body": [{"key": "jon", "type": "boolean"}],
        }
        with self.assertRaises(KeyError):
            spos.bin_encode({"jon": True}, payload_spec)

    def test_encode_version_version_overflow(self):
        payload_spec = {
            "name": "john",
            "version": 17,
            "meta": {"encode_version": True, "version_bits": 4},
            "body": [{"key": "jon", "type": "boolean"}],
        }
        with self.assertRaises(ValueError):
            spos.bin_encode({"jon": True}, payload_spec)

    def test_static_header(self):
        payload_spec = {
            "name": "john",
            "version": 1,
            "meta": {
                "encode_version": True,
                "version_bits": 4,
                "header": [{"key": "my key", "value": "hello!"}],
            },
            "body": [{"key": "jon", "type": "boolean"}],
        }
        t = {"jon": True}
        enc = spos.bin_encode(t, payload_spec)
        hex_enc = spos.encode(t, payload_spec, "hex")
        dec = spos.bin_decode(enc, payload_spec)
        dec, dec_meta = dec["body"], dec["meta"]
        self.assertDict(dec, t)
        self.assertDict(
            dec_meta,
            {
                "name": payload_spec["name"],
                "version": payload_spec["version"],
                "message": hex_enc,
                "header": {"my key": "hello!"},
            },
        )

    def test_static_header_key_key_error(self):
        payload_spec = {
            "name": "john",
            "version": 1,
            "meta": {
                "encode_version": True,
                "version_bits": 4,
                "header": [{"value": "hello!"}],
            },
            "body": [{"key": "jon", "type": "boolean"}],
        }
        t = {"jon": True}
        with self.assertRaises(KeyError):
            spos.bin_encode(t, payload_spec)

    def test_static_header_excess_keys_error(self):
        payload_spec = {
            "name": "john",
            "version": 1,
            "meta": {
                "encode_version": True,
                "version_bits": 4,
                "header": [
                    {"key": "my key", "value": "hello!", "error key": 1}
                ],
            },
            "body": [{"key": "jon", "type": "boolean"}],
        }
        t = {"jon": True}
        with self.assertRaises(KeyError):
            spos.bin_encode(t, payload_spec)


class TestEncodeDecode(TestCase):
    def test_get_subitems(self):
        payload_data = {"holy": {"grail": True, "deeper": {"mariana": 11}}}
        payload_spec = {
            "name": "test get subitems",
            "version": 1,
            "body": [
                {"key": "holy.grail", "type": "boolean"},
                {"key": "holy.deeper.mariana", "type": "integer", "bits": 6},
            ],
        }
        encoded = "0b10010110"
        decoded = {"holy": {"grail": True, "deeper": {"mariana": 11}}}
        enc = spos.bin_encode(payload_data, payload_spec)
        hex_enc = spos.encode(payload_data, payload_spec, "hex")
        self.assertEqual(enc, encoded)
        dec = spos.bin_decode(encoded, payload_spec)
        dec, dec_meta = dec["body"], dec["meta"]
        self.assertDict(dec, decoded)
        self.assertDict(
            dec_meta,
            {
                "name": payload_spec["name"],
                "version": payload_spec["version"],
                "message": hex_enc,
            },
        )

    def test_encode_static_value(self):
        payload_data = {"date": 0.221}
        payload_spec = {
            "name": "test encode",
            "version": 1,
            "meta": {"crc8": True},
            "body": [
                {
                    "key": "holy",
                    "type": "string",
                    "value": "abc",
                    "length": 10,
                },
                {"key": "type", "type": "integer", "value": 10, "bits": 6},
                {"key": "date", "type": "float", "bits": 6},
            ],
        }
        encoded = "0b11111011111011111011111011111011111011111001101001101101110000101000111000100100"
        decoded = {
            "holy": "+++++++abc",
            "type": 10,
            "date": 0.221,
        }
        enc = spos.bin_encode(payload_data, payload_spec)
        hex_enc = spos.encode(payload_data, payload_spec, "hex")
        self.assertEqual(enc, encoded)
        dec = spos.bin_decode(encoded, payload_spec)
        dec, dec_meta = dec["body"], dec["meta"]
        self.assertDict(dec, decoded)
        self.assertDict(
            dec_meta,
            {
                "name": payload_spec["name"],
                "version": payload_spec["version"],
                "message": hex_enc,
                "crc8": True,
            },
        )

    def test_bin_encode_decode_0(self):
        payload_data = {"holy": "grail", "buffer": [1, 2, 3, 4], "date": 0.98}
        payload_spec = {
            "name": "test encode",
            "version": 1,
            "meta": {"crc8": True},
            "body": [
                {"key": "holy", "type": "string", "length": 10},
                {"key": "version", "type": "integer", "value": 1, "bits": 6},
                {
                    "key": "buffer",
                    "type": "array",
                    "bits": 8,
                    "blocks": {"key": "buf_val", "type": "integer", "bits": 8},
                },
                {"key": "date", "type": "float", "bits": 6},
            ],
        }
        encoded = "0b111110111110111110111110111110100000101011011010100010100101000001000001000000000100000010000000110000010011111010000100"
        decoded = {
            "holy": "+++++grail",
            "version": 1,
            "buffer": [1, 2, 3, 4],
            "date": 0.98,
        }
        enc = spos.bin_encode(payload_data, payload_spec)
        hex_enc = spos.encode(payload_data, payload_spec, "hex")
        self.assertEqual(enc, encoded)
        dec = spos.bin_decode(encoded, payload_spec)
        dec, dec_meta = dec["body"], dec["meta"]
        self.assertDict(dec, decoded)
        self.assertDict(
            dec_meta,
            {
                "name": payload_spec["name"],
                "version": payload_spec["version"],
                "message": hex_enc,
                "crc8": True,
            },
        )

    def test_bin_encode_decode_1(self):
        payloads = [
            {
                "sent_yesterday": 0,
                "rpi_temperature": 25,
                "voltage": 8,
                "temperature": 23.3,
                "count_arm": 3,
                "count_eri": 5,
                "count_cos": 10,
                "count_fru": 15,
                "count_sac": 20,
            },
            {
                "sent_yesterday": 1,
                "rpi_temperature": 35,
                "voltage": 11.9,
                "temperature": 27.3,
                "count_arm": 10,
                "count_eri": 6,
                "count_cos": 12,
                "count_fru": 18,
                "count_sac": 24,
            },
            {
                "sent_yesterday": 0,
                "rpi_temperature": 51,
                "voltage": 12.5,
                "temperature": 11.8,
                "count_arm": 5,
                "count_eri": 20,
                "count_cos": 6,
                "count_fru": 7,
                "count_sac": 14,
            },
            {
                "sent_yesterday": 1,
                "rpi_temperature": 77,
                "voltage": 14.1,
                "temperature": 34.5,
                "count_arm": 60,
                "count_eri": 8,
                "count_cos": 16,
                "count_fru": 24,
                "count_sac": 32,
            },
            {
                "sent_yesterday": 0,
                "rpi_temperature": 55,
                "voltage": 4.5,
                "temperature": 24.5,
                "count_arm": 11,
                "count_eri": 5,
                "count_cos": 40,
                "count_fru": 3,
                "count_sac": 9,
            },
        ]

        payload_spec = {
            "name": "test payload 1",
            "version": 2,
            "meta": {"crc8": True},
            "body": [
                {"key": "pad", "type": "pad", "bits": 5},
                {
                    "key": "msg_version",
                    "type": "integer",
                    "value": 2,
                    "bits": 6,
                },
                {"key": "sent_yesterday", "type": "boolean"},
                {
                    "key": "rpi_temperature",
                    "type": "steps",
                    "steps": [30, 50, 75],
                    "steps_names": ["T<30", "30<T<50", "50<T<75", "T>75"],
                },
                {
                    "key": "voltage",
                    "type": "float",
                    "bits": 6,
                    "lower": 10,
                    "upper": 13,
                },
                {
                    "key": "temperature",
                    "type": "float",
                    "bits": 6,
                    "lower": 5,
                    "upper": 50,
                },
                {"key": "count_arm", "type": "integer", "bits": 6},
                {"key": "count_eri", "type": "integer", "bits": 6},
                {"key": "count_cos", "type": "integer", "bits": 6},
                {"key": "count_fru", "type": "integer", "bits": 6},
                {"key": "count_sac", "type": "integer", "bits": 6},
            ],
        }

        for payload_data in payloads:
            enc = spos.bin_encode(payload_data, payload_spec)
            hex_enc = spos.encode(payload_data, payload_spec, "hex")
            dec = spos.bin_decode(enc, payload_spec)
            dec, dec_meta = dec["body"], dec["meta"]
            payload_data["sent_yesterday"] = bool(
                payload_data["sent_yesterday"]
            )
            payload_data["msg_version"] = 2
            if (
                payload_data["voltage"] < 10
            ):  # There's an underflow in the data
                payload_data["voltage"] = 10
            del dec["rpi_temperature"], payload_data["rpi_temperature"]
            self.assertDict(dec, payload_data, 3)
            self.assertDict(
                dec_meta,
                {
                    "name": payload_spec["name"],
                    "version": payload_spec["version"],
                    "message": hex_enc,
                    "crc8": True,
                },
            )

    def test_bin_encode_decode_2(self):
        payload_data = {
            "confidences": [0.9, 0.8, 0.7],
            "categories": ["bike", "bike", "scooter"],
            "timestamp": 1234567890,
            "voltage": 12,
            "temperature": 45,
            "rain": 20,
        }
        payload_spec = {
            "name": "test payload 2",
            "version": 2,
            "body": [
                {
                    "key": "confidences",
                    "type": "array",
                    "bits": 8,
                    "blocks": {
                        "key": "confidence",
                        "type": "float",
                        "bits": 4,
                    },
                },
                {
                    "key": "categories",
                    "type": "array",
                    "bits": 8,
                    "blocks": {
                        "key": "category",
                        "type": "categories",
                        "categories": ["bike", "skate", "scooter"],
                    },
                },
                {
                    "key": "msg_version",
                    "type": "integer",
                    "value": 1,
                    "bits": 6,
                },
                {"key": "timestamp", "type": "integer", "bits": 32},
                {
                    "key": "voltage",
                    "type": "float",
                    "bits": 8,
                    "lower": 10,
                    "upper": 13,
                },
                {
                    "key": "temperature",
                    "type": "float",
                    "bits": 8,
                    "lower": 5,
                    "upper": 50,
                },
                {"key": "rain", "type": "steps", "steps": [0, 10, 20, 30]},
            ],
        }
        enc = spos.bin_encode(payload_data, payload_spec)
        hex_enc = spos.encode(payload_data, payload_spec, "hex")
        dec = spos.bin_decode(enc, payload_spec)
        dec, dec_meta = dec["body"], dec["meta"]
        payload_data["msg_version"] = 1
        payload_data["rain"] = "20<=x<30"
        self.assertDict(dec, payload_data, 3)
        self.assertDict(
            dec_meta,
            {
                "name": payload_spec["name"],
                "version": payload_spec["version"],
                "message": hex_enc,
            },
        )

    def test_hex_encode_decode_0(self):
        payload_data = {"holy": "grail", "buffer": [1, 2], "date": 0.98}
        payload_spec = {
            "name": "test encode",
            "version": 1,
            "body": [
                {"key": "holy", "type": "string", "length": 11},
                {
                    "key": "buffer",
                    "type": "array",
                    "bits": 8,
                    "blocks": {"key": "buf_val", "type": "integer", "bits": 9},
                },
                {"key": "date", "type": "float", "bits": 7},
            ],
        }
        decoded = {
            "holy": "++++++grail",
            "buffer": [1, 2],
            "date": 0.98,
        }
        enc = spos.encode(payload_data, payload_spec, "hex")
        dec = spos.decode(enc, payload_spec)
        dec, dec_meta = dec["body"], dec["meta"]
        self.assertDict(dec, decoded)
        self.assertDict(
            dec_meta,
            {
                "name": payload_spec["name"],
                "version": payload_spec["version"],
                "message": enc,
            },
        )

    def test_encode_decode_0(self):
        payload_data = {"holy": "grail", "buffer": [1, 2, 3, 4], "date": 0.98}
        payload_spec = {
            "name": "test encode",
            "version": 1,
            "meta": {"crc8": True},
            "body": [
                {"key": "holy", "type": "string", "length": 10},
                {"key": "version", "type": "integer", "value": 1, "bits": 6},
                {
                    "key": "buffer",
                    "type": "array",
                    "bits": 8,
                    "blocks": {"key": "buf_val", "type": "integer", "bits": 8},
                },
                {"key": "date", "type": "float", "bits": 6},
            ],
        }
        encoded = b"\xfb\xef\xbe\xfa\n\xda\x8aPA\x00@\x80\xc1>\x84"
        decoded = {
            "holy": "+++++grail",
            "version": 1,
            "buffer": [1, 2, 3, 4],
            "date": 0.98,
        }
        enc = spos.encode(payload_data, payload_spec, "bytes")
        hex_enc = spos.encode(payload_data, payload_spec, "hex")
        self.assertEqual(enc, encoded)
        dec = spos.decode(encoded, payload_spec)
        dec, dec_meta = dec["body"], dec["meta"]
        self.assertDict(dec, decoded)
        self.assertDict(
            dec_meta,
            {
                "name": payload_spec["name"],
                "version": payload_spec["version"],
                "message": hex_enc,
                "crc8": True,
            },
        )

    def test_encode_decode_1(self):
        payload_spec = {
            "name": "test encode decode 1",
            "version": 2,
            "body": [
                {"key": "active", "type": "boolean"},
                {"key": "s3cr37", "type": "binary", "bits": 16},
                {"key": "timestamp", "type": "integer", "bits": 32},
                {"key": "wind speed", "type": "float", "bits": 7},
                {"key": "pad", "type": "pad", "bits": 7},
                {
                    "key": "sensor X",
                    "type": "object",
                    "blocklist": [
                        {"key": "value Y", "type": "integer", "bits": 6},
                        {"key": "value Z", "type": "float", "bits": 6},
                    ],
                },
                {"key": "user input", "type": "string", "length": 7},
                {
                    "key": "bird sightings",
                    "type": "steps",
                    "steps": [0, 5, 10, 15, 20],
                    "steps_names": [
                        "Bogey",
                        "Par",
                        "Birdie",
                        "Eagle",
                        "Albatross",
                        "Condor",
                    ],
                },
                {
                    "key": "battery",
                    "type": "categories",
                    "categories": ["critical", "low", "charged", "full"],
                },
                {
                    "key": "occurences",
                    "type": "array",
                    "bits": 6,
                    "blocks": {
                        "key": "species",
                        "type": "object",
                        "blocklist": [
                            {
                                "key": "name",
                                "type": "categories",
                                "categories": ["kitten", "doggo"],
                            },
                            {"key": "count", "type": "integer", "bits": 6},
                        ],
                    },
                },
            ],
        }
        payload_data = {
            "active": True,
            "s3cr37": "0b1011111011101111",
            "timestamp": 1584042831,
            "wind speed": 0.7,
            "sensor X": {"value Y": 10, "value Z": 0.3},
            "user input": "burguer",
            "bird sightings": 1337,
            "battery": "charged",
            "occurences": [
                {"name": "kitten", "count": 4},
                {"name": "doggo", "count": 10},
            ],
        }
        enc = spos.encode(payload_data, payload_spec)
        hex_enc = spos.encode(payload_data, payload_spec, "hex")
        dec = spos.decode(enc, payload_spec)
        dec, dec_meta = dec["body"], dec["meta"]
        payload_data["bird sightings"] = "Condor"
        self.assertDict(dec, payload_data)
        self.assertDict(
            dec_meta,
            {
                "name": payload_spec["name"],
                "version": payload_spec["version"],
                "message": hex_enc,
            },
        )

    def test_decode_unknown_message_error(self):
        payload_spec = {
            "name": "test decode",
            "version": 2,
            "body": [{"key": "temperature", "type": "float", "bits": 10}],
        }
        with self.assertRaises(ValueError):
            spos.decode(["error"], payload_spec)

    def test_decode_string_message_error(self):
        payload_spec = {
            "name": "test decode",
            "version": 2,
            "body": [{"key": "temperature", "type": "float", "bits": 10}],
        }
        with self.assertRaises(ValueError):
            spos.decode("error string", payload_spec)

    def test_encode_decode_version_only(self):
        payload_spec = {
            "name": "test encode decode version only",
            "version": 17,
            "meta": {"encode_version": True, "version_bits": 6, "crc8": True},
            "body": [],
        }
        enc = spos.encode({}, payload_spec)
        hex_enc = spos.encode({}, payload_spec, "hex")
        dec = spos.decode(enc, payload_spec)
        dec, dec_meta = dec["body"], dec["meta"]
        self.assertDict(dec, {})
        self.assertDict(
            dec_meta,
            {
                "name": payload_spec["name"],
                "version": payload_spec["version"],
                "message": hex_enc,
                "crc8": True,
            },
        )


class TestDecodeFromSpecs(TestCase):
    def setUp(self):
        self.payload_spec_0 = {
            "name": "my spec",
            "version": 0,
            "meta": {"encode_version": True, "version_bits": 6},
            "body": [
                {"key": "sensor x", "type": "boolean"},
                {"key": "sensor y", "type": "integer", "bits": 10},
            ],
        }
        self.payload_spec_1 = {
            "name": "my spec",
            "version": 1,
            "meta": {"encode_version": True, "version_bits": 6},
            "body": [
                {"key": "sensor a", "type": "float", "bits": 6},
                {"key": "sensor b", "type": "integer", "bits": 10},
            ],
        }
        self.payload_spec_2 = {
            "name": "my spec",
            "version": 2,
            "meta": {"encode_version": True, "version_bits": 6},
            "body": [
                {"key": "temperature", "type": "float", "bits": 10},
                {"key": "sunlight", "type": "float", "bits": 8},
            ],
        }
        self.payload_spec_3 = {
            "name": "my spec",
            "version": 3,
            "meta": {"encode_version": True, "version_bits": 6},
            "body": [
                {"key": "night", "type": "boolean"},
                {"key": "fog", "type": "boolean"},
            ],
        }
        self.specs = [
            self.payload_spec_0,
            self.payload_spec_1,
            self.payload_spec_2,
            self.payload_spec_3,
        ]

    def test_decode_from_specs(self):
        t = {"sensor x": False, "sensor y": 19}
        enc = spos.bin_encode(t, self.payload_spec_0)
        hex_enc = spos.encode(t, self.payload_spec_0, "hex")
        dec = spos.decode_from_specs(enc, self.specs)
        dec, dec_meta = dec["body"], dec["meta"]
        self.assertDict(dec, t)
        self.assertDict(
            dec_meta,
            {
                "name": self.payload_spec_0["name"],
                "version": self.payload_spec_0["version"],
                "message": hex_enc,
            },
        )

        t = {"sensor a": 0.4, "sensor b": 500}
        enc = spos.bin_encode(t, self.payload_spec_1)
        hex_enc = spos.encode(t, self.payload_spec_1, "hex")
        dec = spos.decode_from_specs(enc, self.specs)
        dec, dec_meta = dec["body"], dec["meta"]
        self.assertDict(dec, t)
        self.assertDict(
            dec_meta,
            {
                "name": self.payload_spec_1["name"],
                "version": self.payload_spec_1["version"],
                "message": hex_enc,
            },
        )

    def test_decode_from_specs_message_version_mismatch(self):
        t = {"sensor x": False, "sensor y": 19}
        enc = spos.bin_encode(t, self.payload_spec_0)
        enc = f"{enc[:2]}1{enc[3:]}"
        with self.assertRaises(spos.PayloadSpecError):
            spos.decode_from_specs(enc, self.specs)

    def test_decode_from_specs_name_mismatch(self):
        self.payload_spec_0["name"] = "not my spec"
        t = {"sensor x": False, "sensor y": 19}
        enc = spos.bin_encode(t, self.payload_spec_0)
        with self.assertRaises(spos.SpecsVersionError):
            spos.decode_from_specs(enc, self.specs)

    def test_decode_from_specs_duplicate_version(self):
        self.payload_spec_3["version"] = 1
        t = {"sensor x": False, "sensor y": 19}
        enc = spos.bin_encode(t, self.payload_spec_0)
        with self.assertRaises(spos.SpecsVersionError):
            spos.decode_from_specs(enc, self.specs)


class TestStats(TestCase):
    def test_stats(self):
        payload_spec = {
            "name": "spec payload",
            "version": 5,
            "meta": {"encode_version": True, "version_bits": 10, "crc8": True},
            "body": [
                {"key": "pad", "type": "pad", "bits": 5},
                {
                    "key": "msg_version",
                    "type": "integer",
                    "value": 2,
                    "bits": 6,
                },
                {"key": "sent_yesterday", "type": "boolean"},
                {
                    "key": "rpi_temperature",
                    "type": "steps",
                    "steps": [30, 50, 75],
                    "steps_names": ["T<30", "30<T<50", "50<T<75", "T>75"],
                },
                {
                    "key": "voltage",
                    "type": "float",
                    "bits": 6,
                    "lower": 10,
                    "upper": 13,
                },
                {
                    "key": "temperature",
                    "type": "float",
                    "bits": 6,
                    "lower": 5,
                    "upper": 50,
                },
                {
                    "key": "conf_arm",
                    "type": "array",
                    "bits": 3,
                    "blocks": {"key": "conf", "type": "float", "bits": 5},
                },
                {
                    "key": "conf_eri",
                    "type": "array",
                    "bits": 3,
                    "blocks": {"key": "conf", "type": "float", "bits": 5},
                },
                {
                    "key": "conf_cos",
                    "type": "array",
                    "bits": 3,
                    "blocks": {"key": "conf", "type": "float", "bits": 5},
                },
                {
                    "key": "conf_fru",
                    "type": "array",
                    "bits": 3,
                    "blocks": {"key": "conf", "type": "float", "bits": 5},
                },
                {
                    "key": "conf_sac",
                    "type": "array",
                    "bits": 3,
                    "blocks": {"key": "conf", "type": "float", "bits": 5},
                },
            ],
        }
        st = spos.stats(payload_spec)
        self.assertEqual(st["max_bits"], 234)
        self.assertEqual(st["min_bits"], 59)



"""
SPOS - Small Payload Object Serializer
MIT License

Copyright (c) 2020 [Luiz Eduardo Amaral](luizamaral306@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import abc
from string import ascii_uppercase, ascii_lowercase, digits
import copy
import math
import re
import warnings

from .utils import truncate_bits, nest_keys
from .exceptions import StaticValueMismatchWarning


# ---------------------------------------------------------------------
# BLOCK ABC
# ---------------------------------------------------------------------
class BlockBase(abc.ABC):
    """
    Block ABC

    The block is the base object for encoding/decoding messages for
    SPOS.

    A block_spec is required for instantiating a block, required and
    optional keys must be described in class attributes with the same
    name.

    To create a block type it's necessary to inherit this abc and
    implement at least the methods `_bin_encode` and `_bin_decode`.

    """

    required = {}
    optional = {}

    def __init__(self, block_spec):
        self.block_spec = copy.deepcopy(block_spec)
        self.validate_block_spec_keys()
        self.key = block_spec.get("key")
        self.type = block_spec.get("type")
        if not hasattr(self, "value"):
            self.value = block_spec.get("value")
        if not hasattr(self, "bits"):
            self.bits = block_spec.get("bits", 0)

        self.initialize_block()
        self.cache_message = None
        self.cache_value = None
        if self.value is not None:
            self.cache_message = self.bin_encode(self.value)
            self.cache_value = self.bin_decode(self.cache_message)

    def __repr__(self):
        return f"block => {self.block_spec}"

    @abc.abstractmethod
    def _bin_encode(self, value):
        """
        Required method to be implemented for subclasses
        """

    def bin_encode(self, value):
        """
        Method for value binary encoding
        """
        if self.cache_message is not None:
            return self.cache_message
        return self._bin_encode(value)

    @abc.abstractmethod
    def _bin_decode(self, message):
        """
        Required method to be implemented for subclasses
        """

    def bin_decode(self, message):
        """
        Method for binary message decoding
        """
        value = self._bin_decode(message)
        if self.cache_value is not None:
            if value != self.cache_value:
                warnings.warn(
                    f"Decoded message and static value don't match {value} != {self.cache_value}",
                    StaticValueMismatchWarning,
                )
            return self.cache_value
        return value

    def initialize_block(self):
        """
        Optional method for initializing block_spec
        """

    def validate_block_spec_keys(self):
        """
        Validates the keys of `block_spec`. Keys must be specified as
        class attributes `required` and `optional`.
        """
        for key in self.required:
            if key not in self.block_spec:
                raise KeyError(
                    f"Required key {key} not found for block {self}"
                )
            if key == "blocks":
                self.blocks = Block(self.block_spec["blocks"])
            elif key == "blocklist":
                self.blocklist = [
                    Block(b_spec) for b_spec in self.block_spec["blocklist"]
                ]
            else:
                validate_type(self.required[key], self.block_spec[key])
                setattr(self, key, self.block_spec[key])

        for key in self.optional:
            if key in self.block_spec:
                validate_type(self.optional[key]["type"], self.block_spec[key])
                setattr(self, key, self.block_spec[key])
            else:
                setattr(self, key, self.optional[key]["default"])

        all_keys = (
            set(self.required)
            | set(self.optional)
            | set(("key", "value", "type"))
        )
        for key in self.block_spec:
            if key not in all_keys:
                raise KeyError(f"Unexpected key {key} found in {self}")

    def consume(self, message):
        """
        Decodes the block data while returning any unused bits.

        Args:
            message (str): Binary string of the message
        Returns:
            value: Decoded value
            message_tail (str): Unused message bits
        """
        bits = self.accumulate_bits(message)
        value = self.bin_decode(message[: bits + 2])
        return value, "0b" + message[bits + 2 :]

    def accumulate_bits(self, message):
        """
        Optional method that should return the number of bits of the
        instance type. This will change only for types that changes
        dinamically.
        """
        return self.bits

    @property
    def max_bits(self):
        """
        Returns the max number of bits
        """
        return self._max_bits()

    def _max_bits(self):
        return self.bits

    @property
    def min_bits(self):
        """
        Returns the min number of bits
        """
        return self._min_bits()

    def _min_bits(self):
        return self.bits

    def stats(self):
        """
        Returns block statistics
        """
        return {
            "key": self.key,
            "type": self.type,
            "max_bits": self.max_bits,
            "min_bits": self.min_bits,
        }


# ---------------------------------------------------------------------
# METHOD DECORATORS
# ---------------------------------------------------------------------
def validate_type(types, value):
    if isinstance(types, type):
        types = (types,)
    for t in types:
        if isinstance(value, t):
            break
    else:
        raise TypeError(f"Unexpected type {type(value)}")


def validate_encode_input_types(*types):
    def _validate_wrapper(fn):
        def _validate_type_inner(self, value):
            validate_type(types, value)
            return fn(self, value)

        return _validate_type_inner

    return _validate_wrapper


# ---------------------------------------------------------------------
# TYPES
# ---------------------------------------------------------------------
class BooleanBlock(BlockBase):
    bits = 1

    @validate_encode_input_types(bool, int)
    def _bin_encode(self, value):
        return "0b1" if value else "0b0"

    def _bin_decode(self, message):
        return message == "0b1"


class BinaryBlock(BlockBase):
    required = {"bits": int}

    @validate_encode_input_types(str)
    def _bin_encode(self, value):
        if not re.match("^(0b[0-1]*|0x[0-9a-fA-F]*)$", value):
            raise ValueError(
                f"Value for block '{self.key}' must be a binary string or an hex string, got {value}."
            )

        bit_str = (
            bin(int(value, 2))
            if value.startswith("0b")
            else bin(int(value, 16))
        )
        return truncate_bits(bit_str, self.bits)

    def _bin_decode(self, message):
        return truncate_bits(message, self.bits)


class IntegerBlock(BlockBase):
    required = {"bits": int}
    optional = {"offset": {"type": int, "default": 0}}
    offset = None  # Just to calm down the linter

    @validate_encode_input_types(int)
    def _bin_encode(self, value):
        value -= self.offset
        bits = self.bits
        overflow = 2 ** bits - 1
        bit_str = bin(min([max([value, 0]), overflow]))
        return truncate_bits(bit_str, bits)

    def _bin_decode(self, message):
        return int(message, 2) + self.offset


class FloatBlock(BlockBase):
    required = {"bits": int}
    optional = {
        "lower": {"type": (int, float), "default": 0},
        "upper": {"type": (int, float), "default": 1},
        "approximation": {
            "type": (str),
            "default": "round",
            "choices": ["round", "floor", "ceil"],
        },
    }
    lower, upper, approximation = None, None, None

    @validate_encode_input_types(int, float)
    def _bin_encode(self, value):
        approx = round
        if self.approximation == "floor":
            approx = math.floor
        elif self.approximation == "ceil":
            approx = math.ceil
        overflow = 2 ** self.bits - 1
        delta = self.upper - self.lower
        value = overflow * (value - self.lower) / delta
        bit_str = bin(approx(min([max([value, 0]), overflow])))
        return truncate_bits(bit_str, self.bits)

    def _bin_decode(self, message):
        delta = self.upper - self.lower
        overflow = 2 ** self.bits - 1
        return int(message, 2) * delta / overflow + self.lower


class PadBlock(BlockBase):
    required = {"bits": int}
    value = True

    def _bin_encode(self, value=None):
        return f"0b{'1' * self.bits}"

    def _bin_decode(self, message):
        return None


class ArrayBlock(BlockBase):
    required = {"bits": int, "blocks": "block"}
    blocks = None

    def initialize_block(self):
        self.max_length = 2 ** self.bits - 1
        self.length_block = IntegerBlock({"bits": self.bits, "offset": 0})

    @validate_encode_input_types(list, tuple, set)
    def _bin_encode(self, value):
        message = ""
        length = min([len(value), self.max_length])
        message += self.length_block.bin_encode(length)
        for i, v in enumerate(value):
            if i == self.max_length:
                break
            message += self.blocks.bin_encode(v)[2:]
        return message

    def _bin_decode(self, message):
        length, message = self.length_block.consume(message)
        values = []
        for _ in range(length):
            v, message = self.blocks.consume(message)
            values.append(v)
        return values

    def accumulate_bits(self, message):
        length, message = self.length_block.consume(message)
        return self.bits + length * self.blocks.accumulate_bits(message)

    def _max_bits(self):
        return self.bits + self.max_length * self.blocks.max_bits


class ObjectBlock(BlockBase):
    required = {"blocklist": "blocklist"}
    blocklist = None

    def get_value(self, obj, key):
        """
        Gets a value from object `obj`. Dot '.' separates nested
        objects.

        Args:
            obj (dict): Object to get value
            key (str): Key to acess object

        Returns:
            value

        Raises:
            KeyError: If can't find key
        """
        if "." in key:
            dot_idx = key.index(".")
            k1, k2 = key[:dot_idx], key[dot_idx + 1 :]
            return self.get_value(obj[k1], k2)
        return obj[key]

    @validate_encode_input_types(dict)
    def _bin_encode(self, value):
        values = [
            self.get_value(value, block.key)
            if block.value is None
            else block.value
            for block in self.blocklist
        ]
        return "0b" + "".join(
            [
                block.bin_encode(v)[2:]
                for v, block in zip(values, self.blocklist)
            ]
        )

    def _bin_decode(self, message):
        obj = {}
        for block in self.blocklist:
            obj[block.key], message = block.consume(message)
        obj = nest_keys(obj)
        return obj

    def accumulate_bits(self, message):
        acc = 0
        for block in self.blocklist:
            bits = block.accumulate_bits(message)
            _, message = block.consume(message)
            acc += bits
        return acc

    def _max_bits(self):
        return sum([block.max_bits for block in self.blocklist])

    def _min_bits(self):
        return sum([block.min_bits for block in self.blocklist])

    def stats(self):
        return {
            "key": self.key,
            "type": self.type,
            "max_bits": self.max_bits,
            "min_bits": self.min_bits,
            "blocklist": [block.stats() for block in self.blocklist],
        }


class StringBlock(BlockBase):
    required = {"length": int}
    optional = {"custom_alphabeth": {"type": (dict), "default": {}}}
    length, custom_alphabeth = None, None

    alphabeth = dict(
        enumerate(ascii_uppercase + ascii_lowercase + digits + "+/")
    )
    rev_alphabeth = {val: key for key, val in alphabeth.items()}

    def initialize_block(self):
        self.letter_bits = 6
        self.letter_block = IntegerBlock(
            {"bits": self.letter_bits, "offset": 0}
        )
        self.bits = self.length * self.letter_bits

    @validate_encode_input_types(str)
    def _bin_encode(self, value):
        message = "0b"
        value = value.rjust(self.length, " ")
        rev_custom_alphabeth = {
            val: key for key, val in self.alphabeth.items()
        }
        rev_space_map = {" ": 62}  # Maps spaces to +
        for letter in value:
            val = rev_custom_alphabeth.get(
                letter,
                rev_space_map.get(letter, self.rev_alphabeth.get(letter, 63)),
            )
            message += self.letter_block.bin_encode(val)[2:]
        return message

    def _bin_decode(self, message):
        alphabeth = self.alphabeth.copy()
        alphabeth.update(self.custom_alphabeth)
        value = ""
        for _ in range(self.length):
            l, message = self.letter_block.consume(message)
            value += alphabeth[l]
        return value


class StepsBlock(BlockBase):
    required = {"steps": list}
    optional = {"steps_names": {"type": list, "default": []}}
    steps, steps_names = None, None

    def initialize_block(self):
        if self.steps != sorted(self.steps):
            raise ValueError(
                f"Steps Block {self.key} must be ordered: {self.steps}"
            )
        self.bits = math.ceil(math.log(len(self.steps) + 1, 2))
        self.steps_block = IntegerBlock({"bits": self.bits, "offset": 0})
        if not self.steps_names:
            self.steps_names = (
                ["x<{0}".format(self.steps[0])]
                + [
                    "{0}<=x<{1}".format(lower, upper)
                    for lower, upper in zip(self.steps, self.steps[1:])
                ]
                + ["x>={0}".format(self.steps[0])]
            )
        if len(self.steps_names) != len(self.steps) + 1:
            raise ValueError(
                f"'steps_names' for block {self.key} has to have length 1 + len(steps)."
            )

    @validate_encode_input_types(int, float)
    def _bin_encode(self, value):
        value = ([value >= s for s in self.steps] + [False]).index(False)
        return self.steps_block.bin_encode(value)

    def _bin_decode(self, message):
        value = self.steps_block.bin_decode(message)
        return (
            self.steps_names[value]
            if value < len(self.steps_names)
            else "error"
        )


class CategoriesBlock(BlockBase):
    required = {"categories": list}
    categories = None

    def initialize_block(self):
        self.categories = self.categories + ["unknown"]
        self.bits = math.ceil(math.log(len(self.categories), 2))
        self.categories_block = IntegerBlock({"bits": self.bits, "offset": 0})

    @validate_encode_input_types(str)
    def _bin_encode(self, value):
        value = (
            len(self.categories) - 1
            if value not in self.categories
            else self.categories.index(value)
        )
        return self.categories_block.bin_encode(value)

    def _bin_decode(self, message):
        value = self.categories_block.bin_decode(message)
        return (
            self.categories[value] if value < len(self.categories) else "error"
        )


# ---------------------------------------------------------------------
# GENERIC BLOCK CLASS
# ---------------------------------------------------------------------
class Block:
    """
    The block is the base object for encoding/decoding messages for
    SPOS.

    This is a generic class to instantiate one payload types listed
    in: https://github.com/luxedo/SPOS#block

    Args:
        block_spec (dict): Block specification.
    """

    BLOCK_TYPES = {
        "boolean": BooleanBlock,
        "binary": BinaryBlock,
        "integer": IntegerBlock,
        "float": FloatBlock,
        "pad": PadBlock,
        "array": ArrayBlock,
        "object": ObjectBlock,
        "string": StringBlock,
        "steps": StepsBlock,
        "categories": CategoriesBlock,
    }
    key, type, value, bits = None, None, None, None

    def __new__(cls, block_spec):
        cls.validate_block_spec(cls, block_spec)
        return cls.BLOCK_TYPES[block_spec["type"]](block_spec)

    @staticmethod
    def validate_block_spec(cls, block_spec):
        """
        Check for general block specification keys: key and type.

        Args:
            block_spec (dict): Block specification.

        Raises:
            KeyError, TypeError, ValueError: When `block_spec` does
                not conforms to block specification. See:
                https://github.com/luxedo/SPOS#block
        """
        if "key" not in block_spec:
            raise KeyError(f"Block '{block_spec}' must have 'key'.")
        key = block_spec["key"]
        if not isinstance(key, str):
            raise TypeError(f"Block '{block_spec}' 'key' must be a string.")

        # Check type
        if "type" not in block_spec:
            raise KeyError(f"Block '{key}' must have 'type'.")
        b_type = block_spec["type"]
        if b_type not in cls.BLOCK_TYPES:
            raise ValueError(
                f"Block '{block_spec}' has an unknown 'type' {b_type}."
            )



"""
SPOS - Small Payload Object Serializer
MIT License

Copyright (c) 2020 [Luiz Eduardo Amaral](luizamaral306@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import re

from .blocks import Block
from .exceptions import (
    VersionError,
    PayloadSpecError,
    SpecsVersionError,
    StaticValueMismatchWarning,
)
from .checks import create_crc8, check_crc8
from . import utils

__version__ = "1.2.4-b"


def encode_block(value, block_spec):
    """
    Encodes value according to block specifications.

    Args:
        value: The value to be encoded
        block_spec (dict): Block specification

    Returns:
        message (str): Binary string of the message.
    """
    return Block(block_spec).bin_encode(value)


def decode_block(message, block_spec):
    """
    Encodes value according to block specifications.

    Args:
        value: The value to be encoded
        block (dict): Block specifications

    Returns:
        value: Value of the message.
    """
    return Block(block_spec).bin_decode(message)


def _build_meta_block(payload_spec):
    """
    Performs validations in the meta specification and returns o

    Args:
        payload_spec (dict): Payload specifications.

    Returns:
        version_block (Block): A block to encode the version if
            'meta.encode_version' is set otherwise returns None.
        header_block (BlocK): A block to encode the header values.
        header_static (dict): Static values declared in the header.
    """
    meta_spec = payload_spec.get("meta", {})
    version_block = (
        None
        if not meta_spec.get("encode_version")
        else Block(
            {
                "key": "version",
                "type": "integer",
                "bits": meta_spec["version_bits"],
            }
        )
    )

    header_bl = meta_spec.get("header", [])
    header_encode = [block for block in header_bl if "value" not in block]
    header_block = Block(
        {"key": "header", "type": "object", "blocklist": header_encode}
    )

    header_static = {
        block["key"]: block["value"] for block in header_bl if "value" in block
    }
    return version_block, header_block, header_static


def bin_encode(payload_data, payload_spec):
    """
    Encodes a message from payload_data according to payload_spec.
    Returns the message as a binary string.

    Args:
        payload_data (dict): Payload data.
        payload_spec (dict): Payload specification.

    Returns:
        message (str): Binary string of the message.
    """
    utils.validate_payload_spec(payload_spec)
    message = "0b"

    version_block, header_block, header_static = _build_meta_block(
        payload_spec
    )

    if version_block:
        message += version_block.bin_encode(payload_spec["version"])[2:]
    message += header_block.bin_encode(payload_data)[2:]

    body = payload_spec.get("body", [])
    body_block = Block({"key": "body", "type": "object", "blocklist": body})
    message += body_block.bin_encode(payload_data)[2:]

    message += "0" * (
        (8 - (len(message[2:]) % 8)) % 8
    )  # pad message to fill a byte

    if payload_spec.get("meta", {}).get("crc8"):
        message += create_crc8(message)[2:]

    return message


def bin_decode(message, payload_spec):
    """
    Decodes a binary message according to payload_spec.

    Args:
        message (str): Binary string of the message.
        payload_spec (dict): Payload specification.

    Returns:
        body (dict): Payload data.
        meta (dict): Payload metadata.
    """
    utils.validate_payload_spec(payload_spec)
    hex_message = _bin_to_hex(message)
    meta = {
        "name": payload_spec["name"],
        "version": payload_spec["version"],
        "message": hex_message,
    }

    version_block, header_block, header_static = _build_meta_block(
        payload_spec
    )

    if payload_spec.get("meta", {}).get("crc8"):
        meta["crc8"] = check_crc8(message)
        message = message[:-8]

    if version_block:
        msg_version, message = version_block.consume(message)
        if msg_version != meta["version"]:
            raise VersionError(
                f"Versions don't match. Expected {meta['version']}, got {msg_version}."
            )

    header, body_msg = header_block.consume(message)
    header.update(header_static)

    if header:
        meta["header"] = utils.remove_null_values(header)

    body = payload_spec.get("body", {})
    body_block = Block({"key": "body", "type": "object", "blocklist": body})
    body, msg_tail = body_block.consume(body_msg)
    body = utils.remove_null_values(body)
    return {"meta": meta, "body": body}


def encode(payload_data, payload_spec, output="bin"):
    """
    Encodes a message from payload_data according to payload_spec.

    Args:
        payload_data (dict): Payload data.
        payload_spec (dict): Payload specification.
        output (str): Return format (bin, hex or bytes). default: "bin".

    Returns:
        message (bytes): Message.
    """
    message = bin_encode(payload_data, payload_spec)
    if output in ("hex", "bytes"):
        message = _bin_to_hex(message)
        message = (
            message
            if output == "hex"
            else bytes(bytearray.fromhex(message[2:]))
        )
    return message


def _bin_to_hex(bin_message):
    """
    Converts message from bin to hex

    Args:
        bin_message (str)
    Returns
        hex_message (str)
    """
    bin_message = (
        bin_message[2:] if bin_message.startswith("0b") else bin_message
    )
    return "0x" + "".join(
        "{0:02x}".format(int(bin_message[8 * i : 8 * (i + 1)], 2))
        for i in range(len(bin_message) // 8)
    )


def _hex_to_bin(hex_message):
    """
    Converts message from hex to bin

    Args:
        hex_message (str)
    Returns
        bin_message (str)
    """
    bits = len(hex_message[2:]) * 4
    bin_message = bin(int(hex_message, 16))[2:]
    return "0b" + bin_message.zfill(bits)


def decode(message, payload_spec):
    """
    Decodes a message according to payload_spec.

    Args:
        message (bin | hex | bytes): Message.
        payload_spec (dict): Payload specification.

    Returns:
        body (dict): Payload data.
        meta (dict): Payload metadata.
    """
    if isinstance(message, str):
        if not (
            re.match("^0b[01]+$", message)
            or re.match("^0x[0-9a-fA-F]+$", message)
        ):
            raise ValueError(
                "String message must be either a binary string (0b) or an hex string (0x)"
            )
    elif not isinstance(message, bytes):
        raise ValueError(
            f"Message must be either str or bytes, got {type(message)}"
        )
    message = message.strip() if isinstance(message, str) else message
    message = f"0x{message.hex()}" if isinstance(message, bytes) else message
    if message.startswith("0x"):
        message = _hex_to_bin(message)
    return bin_decode(message, payload_spec)


def decode_from_specs(message, specs):
    """
    Decodes message from an avaliable pool of payload specificaions by
    matching message version with specification version.

    All the payload specifications must have `meta.encode_version` set
    and also the same value for `meta.version_bits`.

    Raises:
        PayloadSpecError: If message version is not in 'specs'
        SpecsVersionError: If names doesn't match or has duplicate versions
        Other Exceptions: For incorrect payload specification syntax.
            see spos.utils.validate_payload_spec and
            block.Block.validate_block_spec_keys

    Args:
        message (bin | hex | bytes): Message.
        specs (list): List of Payload specifications.

    Returns:
        body (dict): Payload data.
        meta (dict): Payload metadata.
    """
    utils.validate_specs(specs, match_versions=True)
    for payload_spec in specs:
        try:
            return decode(message, payload_spec)
        except VersionError:
            pass
    raise PayloadSpecError("Message does not match any version in 'specs'.")


def stats(payload_spec):
    """
    Calculates statistics for given payload specification

    Args:
        payload_spec (dict): Payload specification.

    Returns:
        stats (dict): Computed statistics.
    """
    utils.validate_payload_spec(payload_spec)
    version_block, header_block, header_static = _build_meta_block(
        payload_spec
    )
    body = payload_spec.get("body", [])
    body_block = Block({"key": "body", "type": "object", "blocklist": body})
    st = [version_block.stats()] if version_block is not None else []
    st += [header_block.stats(), body_block.stats()]
    st += [{"key": "crc8", "type": "crc8", "max_bits": 8, "min_bits": 8}]
    return {
        "name": payload_spec["name"],
        "version": payload_spec["version"],
        "max_bits": sum([s["max_bits"] for s in st]),
        "min_bits": sum([s["min_bits"] for s in st]),
        "message_sections": st,
    }



"""
SPOS - Small Payload Object Serializer
MIT License

Copyright (c) 2020 [Luiz Eduardo Amaral](luizamaral306@gmail.com)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import json
import sys
import tempfile

from spos import command
from spos import random as srandom
from . import TestCase, test_specs


class TestCommand(TestCase):
    test_specs = test_specs

    def test_parser(self):
        for name, payload_spec_file in self.test_specs.items():
            args = command.parse(["-p", payload_spec_file])
            for ps in args.payload_specs:
                ps.close()

    def test_encode_decode(self):
        for name, payload_spec_file in self.test_specs.items():
            with self.subTest(f"{name}_command_encode_decode"):
                with open(payload_spec_file, "r") as fp:
                    payload_spec = json.load(fp)
                self._test_encode_decode(
                    payload_spec_file, payload_spec, "bytes"
                )
                self._test_encode_decode(
                    payload_spec_file, payload_spec, "hex"
                )
                self._test_encode_decode(
                    payload_spec_file, payload_spec, "bin"
                )

    def _test_encode_decode(self, payload_spec_file, payload_spec, fmt):
        message, payload_data = srandom.random_payload(payload_spec, fmt)

        tmp_in = tempfile.NamedTemporaryFile()
        tmp_out = tempfile.NamedTemporaryFile()
        tmp_in.write(json.dumps(payload_data).encode("utf-8"))
        tmp_in.flush()
        args = command.parse(
            [
                "-p",
                payload_spec_file,
                "-i",
                tmp_in.name,
                "-o",
                tmp_out.name,
                "-f",
                fmt,
            ]
        )
        command.main(args)

        tmp_in.close()
        tmp_out.close()

        message = (
            message if isinstance(message, bytes) else message.encode("utf-8")
        )
        tmp_in = tempfile.NamedTemporaryFile()
        tmp_out = tempfile.NamedTemporaryFile()
        tmp_in.write(message)
        tmp_in.flush()
        args = command.parse(
            [
                "-d",
                "-p",
                payload_spec_file,
                "-i",
                tmp_in.name,
                "-o",
                tmp_out.name,
                "-f",
                fmt,
            ]
        )
        command.main(args)

        tmp_in.close()
        tmp_out.close()

    def test_random_encode_decode(self):
        for name, payload_spec_file in self.test_specs.items():
            with self.subTest(f"{name}_command_random_encode_decode"):
                self._test_random_encode_decode(payload_spec_file, "bytes")
                self._test_random_encode_decode(payload_spec_file, "hex")
                self._test_random_encode_decode(payload_spec_file, "bin")

    def _test_random_encode_decode(self, payload_spec_file, fmt):
        tmp_in = tempfile.NamedTemporaryFile()
        tmp_out = tempfile.NamedTemporaryFile()
        args = command.parse(
            [
                "-p",
                payload_spec_file,
                "-i",
                tmp_in.name,
                "-o",
                tmp_out.name,
                "-f",
                fmt,
                "-r",
            ]
        )
        command.main(args)

        tmp_in.close()
        tmp_out.close()

        tmp_in = tempfile.NamedTemporaryFile()
        tmp_out = tempfile.NamedTemporaryFile()
        args = command.parse(
            [
                "-p",
                payload_spec_file,
                "-i",
                tmp_in.name,
                "-o",
                tmp_out.name,
                "-f",
                fmt,
                "-r",
                "-d",
                "-m",
            ]
        )
        command.main(args)

        tmp_in.close()
        tmp_out.close()

        tmp_in = tempfile.NamedTemporaryFile()
        tmp_out = tempfile.NamedTemporaryFile()
        args = command.parse(
            [
                "-p",
                payload_spec_file,
                "-i",
                tmp_in.name,
                "-o",
                tmp_out.name,
                "-f",
                fmt,
                "-I",
            ]
        )
        command.main(args)

        tmp_in.close()
        tmp_out.close()

    def test_no_payload_specs_error(self):
        with self.assertRaises(SystemExit):
            command.parse([])

    def test_encode_with_multiple_specs_error(self):
        with self.assertRaises(SystemExit):
            args = command.parse(["-p"] + list(self.test_specs.values()))
            command.main(args)

    def test_payload_stats(self):
        args = command.parse(["-s", "-p"] + list(self.test_specs.values()))
        command.main(args)
